@@ -14,6 +14,7 @@ from fnmatch import fnmatch
 from . import jsonutils
 from datetime import datetime, date
 import json
+import logging
 try:
     import pyodbc
 except ModuleNotFoundError:
@@ -32,13 +33,13 @@ class DBText:
         self.iscreated = master_connection is not None
         self.isconnected = False
         self.startrv = ""
-    
+        self.logger = logging.getLogger("dbtext")
         # user master because attach detach
         try:
             self.cnxn = master_connection or self.make_connection("master")
             self.isconnected = True
         except pyodbc.Error as e:
-            print("Unexpected error for db " + database + ":", e)
+            self.logger.error("Unexpected error for db " + database + ":", e)
             raise
         
     def get_create_db_args(self, **kw):
@@ -53,30 +54,35 @@ class DBText:
             attachsql = "CREATE DATABASE " + self.quote(self.database_name) + self.get_create_db_args(**kw) + ";"
             self.query(attachsql)
         except pyodbc.Error as e:
-            print(f"Unexpected error for create db {self.database_name}:\n{attachsql}\n", e)
+            self.logger.error(f"Unexpected error for create db {self.database_name}:\n{attachsql}\n", e)
             raise
 
     def populate_empty_db(self, sqlfile, tables_dir=None, encoding=None):
         try:
             self.iscreated = True
             with self.make_connection(self.database_name) as ttcxn:
-                if sqlfile and os.path.isfile(sqlfile):
-                    self.read_sql_file(ttcxn, sqlfile, encoding)
+                if sqlfile:
+                    if os.path.isfile(sqlfile):
+                        self.read_sql_file(ttcxn, sqlfile, encoding)
+                    else:
+                        self.logger.warning(f"sqlfile {sqlfile} not found")
 
                 tables_dir = tables_dir or self.get_tables_dir_name()
                 if os.path.isdir(tables_dir):
                     self.read_tables_dir(ttcxn, tables_dir)
+                else:
+                    self.logger.info(f"No data folder found for database {tables_dir}")
 
                 self.readrv(ttcxn)
         except pyodbc.Error as e:
-            print(f"Unexpected error for populate empty db {self.database_name}:\n", e)
+            self.logger.error(f"Unexpected error for populate empty db {self.database_name}:\n", e)
             raise
 
     def execute_setup_query(self, ttcxn, currQuery):
         try:
             ttcxn.cursor().execute(currQuery)
         except pyodbc.Error:
-            print("Failed to execute query:\n" + repr(currQuery))
+            self.logger.error("Failed to execute query:\n" + repr(currQuery))
             raise
             
     def read_sql_file(self, ttcxn, sqlfile, encoding=None):
@@ -105,22 +111,23 @@ class DBText:
         if currQuery.strip():
             self.execute_setup_query(ttcxn, currQuery)
 
-    def read_tables_dir(self, ttcxn, tables_dir_name, verbose=False):
+    def read_tables_dir(self, ttcxn, tables_dir_name):
         failedFiles = []
         for tableFile in glob(os.path.join(tables_dir_name, "*.table")) + glob(os.path.join(tables_dir_name, "*.json")):
             try:
-                if verbose:
-                    print("Reading data from", tableFile)
+                self.logger.debug(f"Loading data from {tableFile}")
                 self.add_table_data(tableFile, ttcxn)
             except pyodbc.IntegrityError as ex:
                 fk_constraint_string = "FOREIGN KEY constraint"
                 if fk_constraint_string in ex.args[1]:
+                    self.logger.debug(f"error when loading data for table {tableFile}: {fk_constraint_string}, will retry this table later")
                     failedFiles.append(tableFile)
                 else:
                     raise ex
                 
         # Things often fail due to constraints, insert everything else and then try them again
         for tableFile in failedFiles:
+            self.logger.debug(f"retrying data insert for table {tableFile}")
             self.add_table_data(tableFile, ttcxn)
               
     @classmethod
@@ -189,9 +196,18 @@ class DBText:
         else:
             return value
     
-    def parse_table_file_to_rowdicts(self, fn):
+    def parse_table_file_to_rowdicts(self, fn, primaryKeys):
         if fn.endswith(".json"):
-            return json.load(open(fn))
+            with open(fn, "r") as f:
+                json_table_data = json.load(f)
+                ids = [self.evaluate_primary_key(primaryKeys, row_data) for row_data in json_table_data]
+                if len(primaryKeys) == 1 and () in ids:
+                    self.logger.info(f"source data file {fn} did not contain primary keys for every record, adding them")
+                    for i, row_data in enumerate(json_table_data):
+                        for pk in primaryKeys:
+                            if not pk in row_data:
+                                row_data[pk] = i
+                return json_table_data
         else:
             tablesDir = os.path.dirname(fn)
             rows = []
@@ -203,8 +219,9 @@ class DBText:
             return rows
         
     def add_table_data(self, fn, ttcxn):
-        rowData = self.parse_table_file_to_rowdicts(fn)
         table_name = os.path.basename(fn).rsplit(".", 1)[0]
+        pkeys = self.get_primary_key_columns(ttcxn, table_name)
+        rowData = self.parse_table_file_to_rowdicts(fn, pkeys)
         for currRowDict in rowData:
             self.insert_row(ttcxn, table_name, currRowDict)
           
@@ -225,13 +242,14 @@ class DBText:
             ttcxn.cursor().execute(sql, *list(data.values()))
         except pyodbc.DatabaseError as e:
             if "Cannot insert explicit value for identity column" in str(e):
+                self.logger.debug("Error when inserting data: 'Cannot insert explicit value for identity column'. "
+                                  "Will retry with IDENTITY_INSERT")
                 return self.insert_row(ttcxn, table_name, data, identity_insert=True)
             elif "conflicted with the FOREIGN KEY constraint" in str(e):
                 raise
             else:
                 from pprint import pformat
-                sys.stderr.write("Failed to insert data into " + table_name + ":\n")
-                sys.stderr.write(pformat(data) + "\n")
+                self.logger.error("Failed to insert data into " + table_name + ":\n" + pformat(data) + "\n")
                 raise
 
     def update_start_rv(self):
@@ -239,7 +257,7 @@ class DBText:
             with self.make_connection(self.database_name) as ttcxn:
                 self.readrv(ttcxn)
         except pyodbc.Error as e:
-            print("Unexpected error for update rv " + self.database_name + ":", e)
+            self.logger.error("Unexpected error for update rv " + self.database_name + ":", e)
             pass
     
     def cursor(self):
@@ -258,7 +276,7 @@ class DBText:
         try:
             self.single()
         except pyodbc.Error:
-            print("Failed to go into single user mode.")
+            self.logger.warning("Failed to go into single user mode.")
         try:
             self.query(q)
         finally:
@@ -277,7 +295,7 @@ class DBText:
                 self.query("DROP DATABASE " + self.database_name + ";")
                 self.iscreated = False
             except pyodbc.Error as e:
-                print("Unexpected error for drop db " + self.database_name + ":", e)
+                self.logger.warning("Unexpected error for drop db " + self.database_name + ":", e)
                
     def get_connection_string(self, driver=True):
         connstr = self.connectionStringTemplate % self.database_name
@@ -401,7 +419,7 @@ class DBText:
         table_data = {}
         with self.make_connection(self.database_name) as ttcxn:
             for tablespec, constraint in subset_data:
-                print("Getting data for table(s)", repr(tablespec), ",", repr(constraint))
+                self.logger.info("Getting data for table(s)", repr(tablespec), ",", repr(constraint))
                 rows, colnames = self.extract_data_for_dump(ttcxn, tablespec, constraint)
                 if len(rows) > 0:
                     self.store_table_data(table_data, tablespec, rows, colnames)
@@ -435,7 +453,7 @@ class DBText:
                     
     def write_all_tables(self, table_file_pattern, blob_pattern, ttcxn, exclude=""):
         for tablename in self.expand_table_names(ttcxn, "*", exclude):
-            print("Making file for table", repr(tablename))
+            self.logger.debug("Making file for table", repr(tablename))
             self.dumptable(ttcxn, tablename, "", table_file_pattern, blob_pattern)
 
     def write_data(self, writeDir, use_master_connection=False, json_format=False, **kw):
@@ -488,11 +506,14 @@ class DBText:
                 self.dumptable(ttcxn, tablename, constraint, table_fn_pattern, blob_patterns, dumpableBlobs)
                 
     def get_primary_key_columns(self, ttcxn, tableName):
-        return tuple([ info[3] for info in ttcxn.cursor().primaryKeys(tableName) ])
+        # Sqlite3 cursor doesn't have 'primaryKeys' attribute
+        if hasattr(ttcxn.cursor(), "primaryKeys"):
+            return tuple([ info[3] for info in ttcxn.cursor().primaryKeys(tableName) ])
+        return ()
         
     def evaluate_primary_key(self, pkeys, row_data):
-        return tuple(row_data[key] for key in pkeys)
-        
+        return tuple(row_data[key] for key in pkeys if key in row_data)
+
     def categorise(self, data1, data2, pkeys):
         created, updated, deleted = [], [], []
         
@@ -530,15 +551,17 @@ class DBText:
 
     def dumpchanges(self, table_fn_pattern, tables_dir=None, exclude=""):
         tables_dir = tables_dir or self.get_tables_dir_name()
+        self.logger.info(f"will dump changes compared with tables_dir {tables_dir}")
         created, updated, deleted = {}, {}, {}
         with self.make_connection(self.database_name) as ttcxn:
             for tableName in self.expand_table_names(ttcxn, "*", exclude):
+                self.logger.debug(f"examining changes in table {tableName}")
                 rows, colinfo = self.extract_data_for_dump(ttcxn, tableName, "")
-                tableFile = os.path.join(tables_dir, tableName + ".json")
-                initial_table_data = self.parse_table_file_to_rowdicts(tableFile) if os.path.isfile(tableFile) else []
+                tableFile = os.path.join(tables_dir, tableName + ".json") # TODO: make this also work with rowdata in other format
                 pkeys = self.get_primary_key_columns(ttcxn, tableName)
-                table_data = self.convert_to_row_dicts(rows, colinfo)
-                c, u, d = self.categorise(initial_table_data, table_data, pkeys)
+                initial_table_data = self.parse_table_file_to_rowdicts(tableFile, pkeys) if os.path.isfile(tableFile) else []
+                final_table_data = self.convert_to_row_dicts(rows, colinfo)
+                c, u, d = self.categorise(initial_table_data, final_table_data, pkeys)
                 if c:
                     created[tableName] = c
                 if u:
